# SYNOPSIS
#
#   AX_COMPILER_VERSION
#
# DESCRIPTION
#
#   Try to determine the version of the current compiler.
#   Result is output in $ax_compiler_version
#
#
# LAST MODIFICATION
#
#   2007-11-08
#
# COPYLEFT
#
#   Copyright (c) 2006 Nicolas Calimet
#
#   Copying and distribution of this file, with or without
#   modification, are permitted in any medium without royalty provided
#   the copyright notice and this notice are preserved.

AC_DEFUN([AX_COMPILER_VERSION],
[
  AC_REQUIRE([AC_PROG_EGREP])

  # get current compiler; don't use $ac_compiler
  eval ax_compiler=`set X $ac_compile; echo $[2]`

  AC_MSG_CHECKING([for $ax_compiler version])

  # check for various version options and remove leading/trailing blanks

  # -dumpversion
  AC_TRY_COMMAND([$ax_compiler -dumpversion < /dev/null 2> /dev/null > conftest.out])
  ax_compiler_version=`cat conftest.out | $EGREP '[[0-9]]'`
  ax_compiler_version=`echo $ax_compiler_version`

  # -v ; convert to lowercase and check for "version"
  if test x"$ax_compiler_version" = x""; then
    AC_TRY_COMMAND([$ax_compiler -v < /dev/null 2>&1 | tr A-Z a-z > conftest.out])
    ax_compiler_version=`cat conftest.out | $EGREP "version" | sed 's,\(.*version[[[:space:]]]*\)\(.*\),\2,'`
    ax_compiler_version=`echo $ax_compiler_version`
  fi

  # -V ; convert to lowercase and check for "version"
  if test x"$ax_compiler_version" = x""; then
    AC_TRY_COMMAND([$ax_compiler -V < /dev/null 2>&1 | tr A-Z a-z > conftest.out])
    ax_compiler_version=`cat conftest.out | $EGREP "version" | sed 's,\(.*version[[[:space:]]]*\)\(.*\),\2,'`
    ax_compiler_version=`echo $ax_compiler_version`
  fi

  # --version ; get only the last word of the first line
  if test x"$ax_compiler_version" = x""; then
    AC_TRY_COMMAND([$ax_compiler --version < /dev/null 2>&1 | head -n 1 > conftest.out])
    ax_compiler_version=`cat conftest.out | 's,.*[[[:space:]]],,'`
    ax_compiler_version=`echo $ax_compiler_version`
  fi

  # print compiler version string
  if test x"$ax_compiler_version" = x""; then
    AC_MSG_RESULT([unkown])
  else
    AC_MSG_RESULT([$ax_compiler_version])
  fi

  rm -f conftest.out
])
